from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcx_camf_a1_message_type = QzssDcrDefinition(
    {
        0: 'Test',
        1: 'Alert',
        2: 'Update',
        3: 'All Clear',
    },
    undefined='Undefined Message Type (Code: %d)'
)


from .qzss_dcr_decoder import QzssDcrDecoder
from .qzss_dcr_decoder_base import QzssDcrDecoderBase
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportBase


class NetQzssDcrDecoder(QzssDcrDecoderBase):
    schema = QzssDcReportBase

    def decode(self):
        if not self.sentence:
            raise EOFError('Encountered EOF')

        self.sentence = self.sentence.strip()

        if len(self.sentence) < 33:
            raise QzssDcrDecoderException(
                'Too Short Sentence',
                self)
        if len(self.sentence) > 33:
            raise QzssDcrDecoderException(
                'Too Long Sentence',
                self)

        # converts the message to bytes type
        try:
            self.message = self.sentence[1:]
        except ValueError:
            raise QzssDcrDecoderException(
                'Invalid Message',
                self)

        # extracts a satellite id
        self.satellite_id = self.sentence[0]
        self.satellite_prn = self.satellite_id | 0x80

        # generates a nmea sentence
        self.nmea = self.message_to_nmea()

        # stacks the next decoder
        return QzssDcrDecoder(**self.get_params()).decode()



from ..decoder import QzssDcrDecoder
from ..decoder import QzssDcrDecoderBase
from ..definition import ublox_qzss_dcr_message_header
from ..definition import ublox_qzss_svid_prn_map
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportBase


class UBloxQzssDcrDecoder(QzssDcrDecoderBase):
    schema = QzssDcReportBase

    def decode(self):
        # extracts a message header, satellite id, and message
        self.message_header = self.sentence[:len(ublox_qzss_dcr_message_header)]

        # checks the message header
        if self.message_header != ublox_qzss_dcr_message_header:
            raise QzssDcrDecoderException(
                f'Unknown Message Header: {self.message_header}',
                self)

        # checks the fletcher's checksum
        sum_a = sum_b = 0
        for b in self.sentence[2: -2]:
            sum_a += b
            sum_b += sum_a
        sum_a &= 0xff
        sum_b &= 0xff
        if sum_a != self.sentence[-2] or sum_b != self.sentence[-1]:
            raise QzssDcrDecoderException(
                f'Checksum Mismatch: expected {sum_a:02X}{sum_b:02X}, but got {self.sentence[-2]:02X}{self.sentence[-1]:02X}',
                self)

        # checks the gnss id
        gnss_id = self.sentence[6]
        if gnss_id != 5:
            raise QzssDcrDecoderException(
                f'This Sentence is not from QZSS: {gnss_id}',
                self)

        # extracts the satellite id
        svid = self.sentence[7]
        self.satellite_prn = ublox_qzss_svid_prn_map[svid]
        self.satellite_id = self.satellite_prn & 0x3f # extracts the lower 6 bits

        # checks the signal id
        sig_id = self.sentence[8]
        if sig_id != 1:
            raise QzssDcrDecoderException(
                f'The Sentence is not an L1S Signal: {sig_id}',
                self)

        # checks the data size
        num_data_word = self.sentence[10]
        if num_data_word * 4 + 8 != len(self.sentence) -(len(ublox_qzss_dcr_message_header) + 2 + 2): # SFRBX + Length + CHK
            raise QzssDcrDecoderException(
                f'Invalid Message Length: {num_data_word}',
                self)

        # extracts the dcr message
        data_offset = 14
        data = b''
        for i in range(num_data_word):
            data += bytes((self.sentence[data_offset + 3 + i * 4],
                           self.sentence[data_offset + 2 + i * 4],
                           self.sentence[data_offset + 1 + i * 4],
                           self.sentence[data_offset + 0 + i * 4]))
        self.message = data[:31] + bytes((data[31] & 0xC0,))

        # generates the nmea sentence
        self.nmea = self.message_to_nmea()

        # stacks the next decoder
        return QzssDcrDecoder(**self.get_params()).decode()



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcx_camf_a10_library_version = QzssDcrDefinition(
    {
        0: '#1',
#        1: '#2',
#        2: '#3',
#        3: '#4',
#        4: '#5',
#        5: '#6',
#        6: '#7',
#        7: '#8',
    },
    undefined='Unsupported Library Version (Code: %d)'
)



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_weather_warning_state = QzssDcrDefinition(
    {
        1: "発表",
        2: "解除",
        # "N*": "発表状況(コード番号：N)",
    },
    undefined="発表状況(コード番号：%d)"
)



from .qzss_dcr_lib.exception import QzssDcrDecoderException
from .qzss_dcr_lib.exception import QzssDcrDecoderNotImplementedError
from .qzss_dcr_lib.interface import decode
from .qzss_dcr_lib.interface import decode_stream
from .qzss_dcr_lib.report import qzss_dc_report



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_weather_related_disaster_sub_category = QzssDcrDefinition(
    {
        1: "暴風雪特別警報",
        2: "大雨特別警報",
        3: "暴風特別警報",
        4: "大雪特別警報",
        5: "波浪特別警報",
        6: "高潮特別警報",
        7: "全ての気象特別警報",
        21: "記録的短時間大雨情報",
        22: "竜巻注意情報",
        23: "土砂災害警戒情報",
        31: "その他の警報等情報要素",
        # "NN*": "警報等情報要素(コード番号：NN)",
    },
    undefined="警報等情報要素(コード番号：%d)"
)



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_typhoon_reference_time_type = QzssDcrDefinition(
    {
        1: "実況",  # Analysis
        2: "推定",  # Estimation
        3: "予報",  # Forecast
    },
    undefined="基点時刻分類(コード番号：%d)"
)



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcx_camf_a6_hazard_onset_week = QzssDcrDefinition(
    {
        0: 'Current',
        1: 'Next',
    },
    undefined='Undefined Hazard Onset (Code: %d)'
)



from ..definition import nmea_qzss_dcr_message_header


def nmea_qzss_dcr_message_extractor(reader, reader_args=(), reader_kwargs={}):
    while True:
        msg = reader(*reader_args, **reader_kwargs)
        if not msg:
            raise EOFError('Encountered EOF')
        elif type(msg) == bytes or type(msg) == bytearray:
            msg = msg.decode()
        if msg.startswith(nmea_qzss_dcr_message_header):
            return msg



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcx_camf_a17_main_subject_for_specific_settings = QzssDcrDefinition(
    {
        0: 'B1 - Improved Resolution of Main Ellipse',
        1: 'B2 - Position of the Centre of the Hazard',
        2: 'B3 - Secondary Ellipse Definition',
        3: 'B4 - Quantitative and detailed information about the Hazard',
    },
    undefined='Unsupported Library Version (Code: %d)'
)



class QzssDcrDecoderException(Exception):
    def __init__(self, message, instance=None):
        self.message = message
        self.instance = instance
        super().__init__(self.message)

    def __str__(self):
        sentence = getattr(self.instance, 'nmea', None) or getattr(self.instance, 'sentence', None)
        if sentence is None:
            return f'{self.message}'
        elif type(sentence) is bytes:
            sentence = "b'" + ''.join(r'\x%02X' % c for c in sentence) + "'"
        elif type(sentence) is not str:
            sentence = str(sentence)
        return f'{self.message} -> {sentence}'


class QzssDcrDecoderNotImplementedError(NotImplementedError):
    def __init__(self, message, instance=None):
        self.message = message
        self.instance = instance
        super().__init__(self.message)

    def __str__(self):
        sentence = getattr(self.instance, 'nmea', None) or getattr(self.instance, 'sentence', None)
        if sentence is None:
            return f'{self.message}'
        elif type(sentence) is bytes:
            sentence = "b'" + ''.join(r'\x%02X' % c for c in sentence) + "'"
        elif type(sentence) is not str:
            sentence = str(sentence)
        return f'{self.message} -> {sentence}'



from datetime import datetime
from datetime import timedelta

from .qzss_dcr_decoder_base import QzssDcrDecoderBase
from ..definition import qzss_dcr_jma_epicenter_and_hypocenter
from ..definition import qzss_dcr_jma_local_government
from ..definition import qzss_dcr_jma_notification_on_disaster_prevention
from ..exception import QzssDcrDecoderException


class QzssDcrDecoderJmaCommon(QzssDcrDecoderBase):
    def extract_day_hour_min_field(self, slider):
        dt_d = self.extract_field(slider, 5)
        if dt_d < 1 or dt_d > 31:
            raise QzssDcrDecoderException(
                f'Invalid Time: {dt_d} as day',
                self)
        dt_h = self.extract_field(slider + 5, 5)
        if dt_h > 23:
            raise QzssDcrDecoderException(
                f'Invalid Time: {dt_h} as hour',
                self)
        dt_mi = self.extract_field(slider + 10, 6)
        if dt_mi > 59:
            raise QzssDcrDecoderException(
                f'Invalid Time: {dt_mi} as minute',
                self)

        dt_y = self.report_time.year
        dt_mo = self.report_time.month
        if dt_d - self.report_time.day > 15:
            if dt_mo == 1:
                dt_mo = 12
                dt_y -= 1
            else:
                dt_mo -= 1
        elif self.report_time.day - dt_d > 15:
            if dt_mo == 12:
                dt_mo = 1
                dt_y += 1
            else:
                dt_mo += 1
        if dt_mo == 2 and dt_d == 29:
            while dt_y % 4 != 0 or \
                    (dt_y % 100 == 0 and dt_y % 400 != 0):
                dt_y += 1

        return datetime(year=dt_y,
                        month=dt_mo,
                        day=dt_d,
                        hour=dt_h,
                        minute=dt_mi)

    def extract_local_government(self, slider):
        lg = self.extract_field(slider, 23)
        try:
            return qzss_dcr_jma_local_government[lg], lg
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Local Government: {lg}',
                self)

    def extract_notification_on_disaster_prevention_fields(self, slider):
        notifications = []
        cos = []
        for i in range(3):
            co = self.extract_field(slider + i * 9, 9)
            if co == 0:
                break
            try:
                notifications.append(qzss_dcr_jma_notification_on_disaster_prevention[co])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Notifications on Disaster Prevention: {co}',
                    self)
            cos.append(co)
        return notifications, cos

    def extract_lat_lon_field(self, slider):
        lat_ns = self.extract_field(slider, 1)
        lat_d = self.extract_field(slider + 1, 7)
        if lat_d > 89:
            raise QzssDcrDecoderException(
                f'Invalid Latitude: {lat_d} as degree',
                self)
        lat_m = self.extract_field(slider + 8, 6)
        if lat_m > 59:
            raise QzssDcrDecoderException(
                f'Invalid Latitude: {lat_m} as minute',
                self)
        lat_s = self.extract_field(slider + 14, 6)
        if lat_s > 59:
            raise QzssDcrDecoderException(
                f'Invalid Latitude: {lat_s} as second',
                self)

        lon_ew = self.extract_field(slider + 20, 1)
        lon_d = self.extract_field(slider + 21, 8)
        if lon_d > 179:
            raise QzssDcrDecoderException(
                f'Invalid Longitude: {lon_d} as degree',
                self)
        lon_m = self.extract_field(slider + 29, 6)
        if lon_m > 59:
            raise QzssDcrDecoderException(
                f'Invalid Longitude: {lon_d} as minute',
                self)
        lon_s = self.extract_field(slider + 35, 6)
        if lon_s > 59:
            raise QzssDcrDecoderException(
                f'Invalid Longitude: {lon_s} as second',
                self)

        return {'lat_ns': lat_ns, 'lat_d': lat_d, 'lat_m': lat_m, 'lat_s': lat_s,
                'lon_ew': lon_ew, 'lon_d': lon_d, 'lon_m': lon_m, 'lon_s': lon_s}

    def extract_depth_field(self, slider):
        de = self.extract_field(slider, 9)
        if de == 501:
            return '500kmより深い', de
        elif de == 511:
            return '不明', de
        elif 501 < de < 511:
            raise QzssDcrDecoderException(
                f'Invalid Depth of Hypocenter: {de}',
                self)
        else:
            return f'{de}km', de

    def extract_magnitude_field(self, slider):
        ma = self.extract_field(slider, 7)
        if ma == 101:
            return '10.0より大きい', ma
        elif ma == 126:
            return '不明(8.0より大きい)', ma
        elif ma == 127:
            return '不明', ma
        elif ma < 1 or (101 < ma < 126):
            raise QzssDcrDecoderException(
                f'Invalid Magnitude: {ma / 10}',
                self)
        else:
            return f'{ma / 10}', ma

    def extract_seismic_epicenter_field(self, slider):
        ep = self.extract_field(slider, 10)
        try:
            return qzss_dcr_jma_epicenter_and_hypocenter[ep], ep
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Seismic Epicenter: {ep}',
                self)

    def extract_expected_tsunami_arrival_time(self, slider):
        ta_h = self.extract_field(slider + 1, 5)
        if ta_h == 31:
            return None
        elif ta_h > 23:
            raise QzssDcrDecoderException(
                f'Invalid JMA Expected Tsunami Arrivale Time: {ta_h} as hour',
                self)

        ta_m = self.extract_field(slider + 6, 6)
        if ta_m == 63:
            return None
        elif ta_m > 59:
            raise QzssDcrDecoderException(
                f'Invalid JMA Expected Tsunami Arrivale Time: {ta_m} as minute',
                self)

        ta_dt = self.report_time + timedelta(self.extract_field(slider, 1))

        return datetime(year=ta_dt.year,
                        month=ta_dt.month,
                        day=ta_dt.day,
                        hour=ta_h,
                        minute=ta_m)



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_information_type = QzssDcrDefinition(
    {
        0: '発表',
        1: '訂正',
        2: '取消',
    },
    undefined="情報形態(コード番号：%d)"
)

qzss_dcr_jma_information_type_en = QzssDcrDefinition(
    {
        0: 'Issue',
        1: 'Correction',
        2: 'Cancellation',
    },
    undefined="Undefined Information Type (Code: %d)"
)



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_prefecture
from ..definition import qzss_dcr_jma_seismic_intensity
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaSeismicIntensity


class QzssDcrDecoderJmaSeismicIntensity(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        self.occurrence_time_of_earthquake = self.extract_day_hour_min_field(53)
        self.seismic_intensities = []
        self.seismic_intensities_raw = []
        self.prefectures = []
        self.prefectures_raw = []
        for i in range(16):
            offset = 69 + i * 9
            es = self.extract_field(offset, 3)
            pl = self.extract_field(offset + 3, 6)
            if es == 0 and pl == 0:
                break

            try:
                self.seismic_intensities.append(qzss_dcr_jma_seismic_intensity[es])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Seismic Intensity: {es}',
                    self)
            self.seismic_intensities_raw.append(es)

            try:
                self.prefectures.append(qzss_dcr_jma_prefecture[pl])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Prefecture: {pl}',
                    self)
            self.prefectures_raw.append(pl)

        return QzssDcReportJmaSeismicIntensity(**self.get_params())



import argparse
import datetime
import sys
from pprint import pformat

from azarashi import QzssDcrDecoderException
from azarashi import QzssDcrDecoderNotImplementedError
from azarashi import decode_stream


def main():
    parser = argparse.ArgumentParser(description='azarashi CLI', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('type', help='message type', type=str, choices=['hex', 'nmea', 'ublox'])
    parser.add_argument('-f', '--input', help='input device', type=str, default='stdin')
    parser.add_argument('-s', '--source', help='output the source messages', action='store_true')
    parser.add_argument('-u', '--unique', help='supress duplicate messages', action='store_true')
    parser.add_argument('-r', '--ignore-dcr', help='ignore dcr messages', action='store_true')
    parser.add_argument('-x', '--ignore-dcx', help='ignore dcr messages', action='store_true')
    parser.add_argument('-v', '--verbose', help="verbose mode", action='store_true')
    args = parser.parse_args()
    if args.input == 'stdin':
        stream = sys.stdin
    else:
        stream = open(args.input, mode='r')

    while True:
        now = datetime.datetime.now().isoformat()
        try:
            report = decode_stream(stream, args.type,
                                   unique=args.unique,
                                   ignore_dcr=args.ignore_dcr,
                                   ignore_dcx=args.ignore_dcx)
            if args.verbose is True:
                print(f'{now} --------------------------------\n{pformat(report.get_params())}\n')
            else:
                print(f'{now} --------------------------------\n{report}\n')

            if args.source is True:
                if type(report.sentence) is bytes:
                    src = "b'" + ''.join(r'\x%02X' % c for c in report.sentence) + "'"
                else:
                    src = report.sentence
                print(f'# src: {src}\n# hex: {report.message.hex().upper()[:-1]}\n')

            sys.stdout.flush()
        except QzssDcrDecoderException as e:
            print(f'{now} --------------------------------\n# [{type(e).__name__}] {e}\n', file=sys.stderr)
        except QzssDcrDecoderNotImplementedError as e:
            print(f'{now} --------------------------------\n# [{type(e).__name__}] {e}\n', file=sys.stderr)
        except EOFError as e:
            print(f'{e}\n', file=sys.stderr)
            stream.close()
            return 0
        except Exception as e:
            print(f'{now} --------------------------------\n# [{type(e).__name__}] {e}\n', file=sys.stderr)
            stream.close()
            return 1


if __name__ == '__main__':
    exit(main())



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_ash_fall_warning_code
from ..definition import qzss_dcr_jma_volcano_name
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaAshFall
from ..report import QzssDcReportJmaBase


class QzssDcrDecoderJmaAshFall(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        self.activity_time = self.extract_day_hour_min_field(53)

        dw1 = self.extract_field(69, 2)
        if dw1 == 1:
            self.ash_fall_warning_type = '速報'
        elif dw1 == 2:
            self.ash_fall_warning_type = '詳細'
        else:
            raise QzssDcrDecoderException(
                f'Undefined JMA Ash Fall Warning Type: {dw1}',
                self)
        self.ash_fall_warning_type_raw = dw1

        vo = self.extract_field(71, 12)
        try:
            self.volcano_name = qzss_dcr_jma_volcano_name[vo]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Volcano Name: {vo}',
                self)
        self.volcano_name_raw = vo

        self.expected_ash_fall_times = []
        self.ash_fall_warning_codes = []
        self.ash_fall_warning_codes_raw = []
        self.local_governments = []
        self.local_governments_raw = []
        for i in range(4):
            offset = 83 + i * 29
            if self.extract_field(offset, 29) == 0:
                break

            ho = self.extract_field(offset, 3)
            if 1 > ho > 6:
                raise QzssDcrDecoderException(
                    f'Invalid JMA Expected Ash Fall Time: {ho}',
                    self)
            self.expected_ash_fall_times.append(ho)

            dw2 = self.extract_field(offset + 3, 3)
            try:
                self.ash_fall_warning_codes.append(qzss_dcr_jma_ash_fall_warning_code[dw2])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Ash Fall Warning Code: {dw2}',
                    self)
            self.ash_fall_warning_codes_raw.append(dw2)

            local_government, lg = self.extract_local_government(offset + 6)
            self.local_governments.append(local_government)
            self.local_governments_raw.append(lg)

        return QzssDcReportJmaAshFall(**self.get_params())



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_northwest_pacific_tsunami_height_en = QzssDcrDefinition(
    {
        1: "0.3m~1m",
        2: "1m~3m",
        3: "3m~5m",
        4: "5m~10m",
        508: "More than 10m",
        509: "Huge",
        510: "High",
        511: "Unknown",
    },
    undefined="Undefined Tsunami Height (Code: %d)"
)
